funcdef void SliderChange(float Value);

class IGSlider {
    // center point
    vec2 Position { get { return vec2(Setting_OSD_Pos_X, Setting_OSD_Pos_Y); } }
    vec2 Size { get { return vec2(Setting_OSD_Pos_W, Setting_OSD_Pos_H); } }
    vec2 ButtonSize { get { return vec2(Setting_OSD_ButtWidth, Setting_OSD_ButtHeight); } }
    float FontSize { get { return Setting_OSD_FontSize; } }

    int Font = 0;

    private bool _visible = true;
    bool Visible {
        get { return _visible; }
        set {
            _visible = value;
            if (!_visible) {
                IsMouseDown = false;
                IsMouseOver = false;
            }
        }
    }

    float BorderWidth = 3.0;
    float BorderRadius = 5.0;
    vec4 BorderColor(1, 1, 1, 0.75);
    vec4 BackColor(0.9, 0.9, 0.9, 0.25);
    vec4 ButtonColor(0.66, 0.66, 0.66, 0.9);
    vec4 TextColor(1, 1, 1, 0.9);
    vec4 TextShadowColor(0,0,0,0.7);
    vec4 MarkerColor(1, 0.66, 0.66, 0.75);
    float ShadowBlur = 1;

    float Min = -1.0;
    float Max = 1.0;

    private float _value = 0;
    float Value {
        get {
            return _value;
        }
        set {
            if (IsMouseDown) { return; }
            _value = value;
        }
    }

    float MarkerValue = -1.0;

    string ButtonText = "";

    // valueText above button format
    string ValueFormat = "%.2f";

    bool DisplayRange = false;

    SliderChange@ OnChange = null;

    IGSlider() { InitFont(); }

    void InitFont() {
        Font = nvg::LoadFont("DroidSans-Bold.ttf");
        auto iconFont = nvg::LoadFont("ManiaIcons.ttf");
        if (iconFont == 0) {
            iconFont = nvg::LoadFont("fa-solid.ttf");
        }
        nvg::AddFallbackFont(Font, iconFont);
    }

    void Changed() {
        if (OnChange !is null) { OnChange(Value); }
    }

    private vec2 DEFAULT_SCREEN(1920, 1080);
    private float DEFAULT_SHADOW_OFFSET = 2;
    private float DEFAULT_TEXT_WIDTH = 300;
    private float DEFAULT_MARKER_WIDTH = 0;

    vec2 ScreenPosition {
        get {
            auto W = Draw::GetWidth();
            auto H = Draw::GetHeight();

            return vec2(W * Position.x, H * Position.y);
        }
    }

    vec2 ScreenSize {
        get {
            float SW = Draw::GetWidth() / DEFAULT_SCREEN.x;
            float SH = Draw::GetHeight() / DEFAULT_SCREEN.y;

            float w = Size.x * SW;
            float h = Size.y * SH;
            return vec2(w, h);
        }
    }

    vec2 ScreenPositionTopLeft {
        get {
            auto p = ScreenPosition;
            auto s = ScreenSize;

            auto hw = s.x / 2.f;
            auto hh = s.y / 2.f;

            return vec2(p.x - hw, p.y - hh);
        }
    }

    vec2 ScreenButtonSize {
        get {
            float SW = Draw::GetWidth() / DEFAULT_SCREEN.x;
            float SH = Draw::GetHeight() / DEFAULT_SCREEN.y;

            float w = ButtonSize.x * SW;
            float h = ButtonSize.y * SH;

            return vec2(w, h);
        }
    }

    vec4 BarBounds {
        get {
            auto p = ScreenPositionTopLeft;
            auto s = ScreenSize;

            return vec4(p.x, p.y, p.x + s.x, p.y + s.y);
        }
    }

    vec4 ButtonBounds {
        get {
            auto p = ScreenPositionTopLeft;
            auto s = ScreenSize;
            auto bs = ScreenButtonSize;

            auto hw = bs.x / 2.0;
            auto hh = bs.y / 2.0;

            float rp = 0;
            if (Max - Min != 0) {
                rp = Math::Clamp((Value - Min) / (Max - Min), 0.f, 1.f);
            }
            float x = p.x + s.x * rp - hw;
            float y = p.y - (bs.y - s.y) / 2.0;

            return vec4(x, y, x + bs.x, y + bs.y);
        }
    }

    vec4 MarkerBounds {
        get {
            auto p = ScreenPositionTopLeft;
            auto s = ScreenSize;

            float SW = Draw::GetWidth() / DEFAULT_SCREEN.x;
            float SH = Draw::GetHeight() / DEFAULT_SCREEN.y;

            auto mw = DEFAULT_MARKER_WIDTH * SW;
            auto mh = s.y;

            float rp = 0;
            if (Max - Min != 0) {
                rp = Math::Clamp((MarkerValue - Min) / (Max - Min), 0.f, 1.f);
            }
            float x = p.x + s.x * rp - (mw / 2.0);
            float y = p.y;

            return vec4(x, y, x + mw, y + mh);
        }
    }

    void DrawShadowedText(const string&in valueText, float x, float y, float w) {
        const float SW = Draw::GetWidth() / DEFAULT_SCREEN.x;
        const float SH = Draw::GetHeight() / DEFAULT_SCREEN.y;

        nvg::BeginPath();
        nvg::FillColor(TextShadowColor);
        nvg::FontBlur(ShadowBlur);
        nvg::TextBox(x + DEFAULT_SHADOW_OFFSET * SW, y + DEFAULT_SHADOW_OFFSET * SH, w, valueText);

        nvg::BeginPath();
        nvg::FillColor(TextColor);
        nvg::FontBlur(0);
        nvg::TextBox(x, y, w, valueText);
    }

    string Format(float value) {
        if (ValueFormat.Length != 0) { // bleh
            return Text::Format(ValueFormat, value);
        } else {
            return Time::Format(int(value * 1000.0));
        }
    }

    void Render() {
        if (!Visible || !IsInterfaceVisible()) {
            return;
        }

        auto bounds = BarBounds;
        const float SW = Draw::GetWidth() / DEFAULT_SCREEN.x;
        const float SH = Draw::GetHeight() / DEFAULT_SCREEN.y;

        nvg::BeginPath();
        nvg::RoundedRect(bounds.x, bounds.y, bounds.z - bounds.x, bounds.w - bounds.y, BorderRadius * SW);
        nvg::ClosePath();

        nvg::FillColor(IsActive ? ChangeBrightness(BackColor, 0.25) : BackColor);
        nvg::Fill();
        nvg::StrokeWidth(BorderWidth * SW);
        nvg::StrokeColor(BorderColor);
        nvg::Stroke();

        if (MarkerValue > Min && MarkerValue < Max) {
            auto markerBounds = MarkerBounds;
            nvg::BeginPath();
            nvg::RoundedRect(markerBounds.x, markerBounds.y, markerBounds.z - markerBounds.x, markerBounds.w - markerBounds.y, BorderRadius * SW);
            nvg::ClosePath();

            nvg::StrokeWidth(BorderWidth * SW);
            nvg::StrokeColor(MarkerColor);
            nvg::Stroke();
        }

        auto buttonBounds = ButtonBounds;
        float textW = DEFAULT_TEXT_WIDTH * SW;
                auto RealFontSize = FontSize * SH;
        if (Font != 0) {
            nvg::FontFace(Font);
        }
        nvg::FontSize(RealFontSize);
        nvg::TextAlign(nvg::Align::Middle | nvg::Align::Center);

        if (DisplayRange) {
            auto rangeTextY = buttonBounds.w + BorderWidth * SH * 2;

            auto startTextX = bounds.x - textW / 2.0;
            auto endTextX = bounds.z - textW / 2.0;

            string startText = Format(Min);
            string endText = Format(Max);

            DrawShadowedText(startText, startTextX, rangeTextY, textW);
            DrawShadowedText(endText, endTextX, rangeTextY, textW);
        }

        nvg::BeginPath();
        // Round because text drawing doesn't do floaty stuff
        nvg::RoundedRect(Math::Round(buttonBounds.x), Math::Round(buttonBounds.y), Math::Round(buttonBounds.z - buttonBounds.x), Math::Round(buttonBounds.w - buttonBounds.y), BorderRadius * SW);
        nvg::ClosePath();

        nvg::FillColor(IsActive ? ChangeBrightness(ButtonColor, 0.4) : ButtonColor);
        nvg::Fill();
        nvg::StrokeWidth(BorderWidth * SW);
        nvg::StrokeColor(BorderColor);
        nvg::Stroke();

        // Round because text drawing doesn't do floaty stuff well
        float buttonTextX = Math::Round(buttonBounds.x + (buttonBounds.z - buttonBounds.x) / 2.0 - DEFAULT_TEXT_WIDTH * SW / 2.0);
        float buttonTextY = Math::Round(buttonBounds.y + (buttonBounds.w - buttonBounds.y) / 2.0);

        string valueText;
        float valueTextX;
        if (IsMouseOver) {
            // display value of position under cursor on mouseover
            auto mpr = Math::Clamp((MousePosition.x - bounds.x) / (bounds.z - bounds.x), 0.0, 1.0);
            valueText = Format(Min + (Max - Min) * mpr);
            valueTextX = bounds.x + (bounds.z - bounds.x) * mpr - textW / 2.0;
        } else {
            // display value over button
            valueText = Format(Value);
            valueTextX = buttonTextX;
        }
        auto valueTextY = buttonBounds.y - RealFontSize / 2.0 - BorderWidth * SH;

        DrawShadowedText(valueText, valueTextX, valueTextY, textW);

        if (ButtonText.Length > 0) {
            DrawShadowedText(ButtonText, buttonTextX, buttonTextY, textW);
        }
    }

    bool IsActive {
        get {
            return IsMouseOver || IsMouseDown;
        }
    }
    bool IsMouseOver = false;
    bool IsMouseDown = false;
    int2 MousePosition;

    bool PointInsideBar(int x, int y) {
        auto bounds = BarBounds;
        return (x >= bounds.x && x <= bounds.z && y >= bounds.y && y <= bounds.w);
    }

    bool PointInsideButton(int x, int y) {
        auto buttonBounds = ButtonBounds;
        return (x >= buttonBounds.x && x <= buttonBounds.z && y >= buttonBounds.y && y <= buttonBounds.w);
    }

    bool PointInside(int x, int y) {
        return PointInsideBar(x, y) || PointInsideButton(x, y);
    }

    void StartSlider(int x, int y) {
        UpdateSlider(x, y);
    }

    void UpdateSlider(int x, int y) {
        auto w = (x - ScreenPositionTopLeft.x) / ScreenSize.x;
        float dv = w * (Max - Min) + Min;
        _value = Math::Clamp(dv, Min, Max);
        Changed();
    }

    void OnMouseMove(int x, int y) {
        if (!Visible) { return; }
        IsMouseOver = PointInside(x, y);
        MousePosition.x = x;
        MousePosition.y = y;
        if (IsMouseDown) {
            UpdateSlider(x, y);
        }
    }

    bool OnMouseButton(bool down, int button, int x, int y) {
        if (!Visible) { return false; }
        IsMouseOver = PointInside(x, y);
        if (IsMouseDown && !down) {
            IsMouseDown = false;
            UpdateSlider(x, y);
            return true;
        } else if (IsMouseOver && down) {
            IsMouseDown = true;
            StartSlider(x, y);
            return true;
        }

        return false;
    }

    void Update(float dt) {
        if (IsMouseDown) {
            Changed();
        }
    }
}
